#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "dislib.h"
#include "symbol.h"
#include "loader.h"

struct opcode_type {
   char name[8];
   int type;
} opcodes[8] = {{"add ",RRRTYPE},
                {"addi",RRITYPE},
                {"nand",RRRTYPE},
                {"lui ",RITYPE},
                {"sw  ",RRITYPE},
                {"lw  ",RRITYPE},
                {"beq ",RRITYPE},
                {"jalr",RRETYPE}};


    /* Split Instructions Into Components */
void split_instruction(unsigned short instruction,short *opcode,
		       short *arg1, short *arg2, short *arg3,
		       short *arg4) {

   
    *opcode=instruction>>OP_SHIFT;
   
    switch(opcodes[*opcode].type) {
       /* OOO111222XXXX333             */
       /* FEDCBA9876543210             */
     case RRRTYPE:
          *arg1=(instruction>>A_SHIFT)&0x7;
          *arg2=(instruction>>B_SHIFT)&0x7;
          *arg3=(instruction&0x7);
          return;
       /* OOO111222IIIIIII             */
       /* FEDCBA9876543210             */       
     case RRITYPE:
          *arg1=(instruction>>A_SHIFT)&0x7;
          *arg2=(instruction>>B_SHIFT)&0x7;
          *arg3=(instruction)&0x7f;
          if (*arg3>63) *arg3=*arg3-128;       
          return;
       /* OOO111222TTT####             */
       /* FEDCBA9876543210             */
     case RRETYPE:
          *arg1=(instruction>>A_SHIFT)&0x7;
          *arg2=(instruction>>B_SHIFT)&0x7;
          *arg3=(instruction>>EXT_SHIFT)&0x7;
          *arg4=(instruction)&0xf;
          return;
       /* OOO111IIIIIIIIII             */
       /* FEDCBA9876543210             */       
     case RITYPE:
          *arg1=(instruction>>A_SHIFT)&0x7;
          *arg2=((instruction)&0x03ff);
          return;
    }
   
}

    /* Dissasemble Machine Code into a string */
void decode_assembly(short assembly,short pc,char *return_string) {

short opcode,arg1,arg2,arg3,arg4;
   
    split_instruction(assembly,&opcode,&arg1,&arg2,&arg3,&arg4);
    if (opcodes[opcode].type==RRRTYPE) {
       sprintf(return_string,"%s\t%i,%i,%i",opcodes[opcode].name,
	                                    arg1,arg2,arg3);
    }
    if (opcodes[opcode].type==RRITYPE) {
       if (opcode==BEQ) {
	  int i;
	  i=find_symbol_by_address(pc+arg3+1,SYMBOL_TEXT);
	  if (i<0) {
	     sprintf(return_string,"%s\t%i,%i,%i",opcodes[opcode].name,
	                                    arg1,arg2,arg3);
	  }
	  else {
	     sprintf(return_string,"%s\t%i,%i,%s",opcodes[opcode].name,
	                                    arg1,arg2,
		     symbol[i].label);	     
	  }
       }
       else {
          sprintf(return_string,"%s\t%i,%i,%i",opcodes[opcode].name,
	                                    arg1,arg2,arg3);
       }
    }
    if ((opcodes[opcode].type==RITYPE)) {
       sprintf(return_string,"%s\t%i,%i",opcodes[opcode].name,
	       arg1,arg2<<6);
    }
    if ((opcodes[opcode].type==RRETYPE)) {
       
       if ((arg1==0) && (arg2==0)) {
	  if (arg3==EXT_EXCEPTION) {  
	     if (arg4==EXC_HALT) {
		sprintf(return_string,"halt");
	     }
	     else {
		sprintf(return_string,"exc\t%i",arg4);
	     }
	  }
	  else if (arg3==EXT_SYSCALL) {
	     sprintf(return_string,"sys\t%i",arg4);
	  }
	  else {
	     sprintf(return_string,"%s\t%i,%i",opcodes[opcode].name,
	       arg1,arg2);   
	  }
       }
       else {
	  sprintf(return_string,"%s\t%i,%i",opcodes[opcode].name,
	       arg1,arg2);   
       }
    }
}


int get_opcode(char *string) {
   
   if (string==NULL) return INVALID;

   if (!strcmp(string, "add")) {
      return ADD;
   }   
   if (!strcmp(string, "addi")) {
      return ADDI;
   }
   if (!strcmp(string, "nand")) {
      return NAND;
   }
   if (!strcmp(string, "lui")) {
      return LUI;
   }
   if (!strcmp(string, "lw")) {
      return LW;
   }
   if (!strcmp(string, "sw")) {
      return SW;
   }
   if (!strcmp(string, "jalr")) {
      return JALR;
   }
   if (!strcmp(string, "beq")) {
      return BEQ;
   }
   if (!strcmp(string, "nop")) {
      return NOP;
   }
   if (!strcmp(string, "halt")) {   
      return HALT;
   }
   if (!strcmp(string, "sys")) {
      return SYS;
   }
   if (!strcmp(string, "exc")) {
      return EXC;
   }
   if (!strcmp(string, "lli")) {
      return LLI;
   }
   if (!strcmp(string, "movi")) {
      return MOVI;
   }
   if (!strcmp(string, ".fill")) {
      return FILL;
   }
   if (!strcmp(string, ".space")) {
      return SPACE;
   }
   if (!strcmp(string, ".include")) {
      return INCLUDE;
   }   
   if (!strcmp(string, ".equ")) {
      return EQU;
   }   
   if (!strcmp(string, ".byte")) {
      return BYTE;
   }   
   if (!strcmp(string, ".ascii")) {
      return ASCII;
   }      

   return INVALID;
}

